/*
 * Bluetooth to ELM327 Interface
 * Uses a HC-05 to connect to a ELM-Interface for OBD II
 * Print values on a 44780 LCD with I²C-adaptor
 * 
 * 29.03.2020 Florian Schäffer  https://www.blafusel.de
 * 
 * Prints on USB-Serial (Baudrate: 115200) debug information
 * Use software serial for HC05
 * Enter RX and TX and Kex/Enable pin
 * Enter Password for ELM
 * Enter MAC address for ELM
 * 
 * China cheap ELM327 clone:    +INQ:D:18:1,1F00,7FFF           =>  D,18,1
 * OBDLink LX:                  +INQ:4:3E:543905,20300,7FFF     =>  4,3E,543905
 * 
 */

#include <LiquidCrystal_I2C.h>    https://github.com/johnrickman/LiquidCrystal_I2C
#include <SoftwareSerial.h>

#define ELMMAC "D,18,1"       // MAC of ELM327
#define ELMPWD "1234"         // BT Pairing PIN for ELM  *** not empty! ***

const uint8_t Pin_RX = 10;
const uint8_t Pin_TX = 11;
const uint8_t Pin_KEY = 9;
char str[50];

LiquidCrystal_I2C lcd(0x27, 16, 2);         // I²C Adress, Cols, Rows
SoftwareSerial BTSerial(Pin_RX, Pin_TX);    // RX | TX

/**
  @brief  Sendet einen String (command) und wartet auf Antwort. 
          Wenn innerhalb 10 Sekunden keine Antwort oder nicht "OK", dann Reset
  @param  command
  @return none
*/
void sendATcmd (char *command)
{
  char recvChar;
  int i=0;
  uint32_t startTime;

  Serial.print(command);
  Serial.print(": ");
  
  BTSerial.write("AT+");
  BTSerial.write(command);
  BTSerial.write("\r\n");

  startTime = millis();
  do              
  {
    if (BTSerial.available())
    {
      recvChar = BTSerial.read();                 //read data from HC-05
      Serial.print(recvChar);
      str[i]=recvChar;                               //put received char to str
      i+=1;
    }
    if (millis() >= startTime+10000)    // wait 10 sec
    {
      Serial.println("\nNO BT DATA");
      delay(500);     // wait for println
      asm("jmp 0x00");      // Reset
    }
  } while (recvChar != '\n');

  if (!((str[0] == 'O') && (str[1] == 'K')))   //if response is !OK: RESET
  {
    Serial.println("\nBT FAIL");
    delay(500);     // wait for println
    asm("jmp 0x00");      // Reset
  }
}

/**
  @brief  Sendet einen String (command) und wartet auf Antwort. 
          Wenn innerhalb 10 Sekunden keine Antwort oder nicht "OK", dann Reset
  @param  command
  @return none
*/
void sendOBDcmd (char *command)
{
  char recvChar;
  int i=0;
  uint32_t startTime;

  Serial.print("->ELM: ");
  Serial.println(command);
  
  BTSerial.write(command);
  BTSerial.write("\r");           // ELM only needs \r

  startTime = millis();
  do              
  {
    if (BTSerial.available())
    {
      recvChar = BTSerial.read();                 //read data from HC-05
      Serial.print(recvChar);
      str[i]=recvChar;                               //put received char to str
      i+=1;
    }
    if (millis() >= startTime+10000)    // wait 15 sec
    {
      Serial.println("\nNO ELM DATA");
      delay(500);     // wait for println is finished
      asm("jmp 0x00");      // Reset
    }
  } while (recvChar != '>');
 
  if ((str[i-1] != '>') || ((i > 4) && ( ((str[i-5] == 'O') && (str[i-4] == 'R')) || ((str[i-5] == 'T') && (str[i-4] == 'A')))))    // if last sign not ">" *or* - but only if String-length > 4: if last 2 signs = "OR" (from "ERROR\r\r>")  *or* last 2 signs = "TA" (from "NO DATA\r\r>") 
  {
    Serial.println("\nELM ERROR");
    delay(500);     // wait for println
    asm("jmp 0x00");      // Reset
  }
 
  Serial.println();
}

/**
  @brief  Initialisiert den HC05
  @param  none
  @return none
*/
void initHC05 ()
{
  Serial.println("Attempting to connect to HC-05...");
  lcd.setCursor(0, 1);
  lcd.print("Init HC-05 ...  ");

  digitalWrite(Pin_KEY, HIGH);
  sendATcmd ("RESET");          // Reset
  delay (1000);
  sendATcmd ("ORGL");           // Restore default
  sendATcmd ("ROLE=1");         // puts the HC-05 in Master Mode
  sendATcmd ("CMODE=0");        // allows the HC-05 to connect to any device
  sendATcmd ("RMAAD");          // Delete All Authenticated Device
  sendATcmd ("PSWD=" ELMPWD);
  sendATcmd ("BIND=" ELMMAC);     
  delay (1000);
  sendATcmd ("INIT");
  delay (1000);
  sendATcmd ("PAIR=" ELMMAC ",20");
  delay (1000);
  sendATcmd ("LINK=" ELMMAC);
  delay (1000);
  digitalWrite(Pin_KEY, LOW);      // HC05: AT-mode OFF => pass through mode

  Serial.println("HC-05 ready");
  lcd.setCursor(0, 1);
  lcd.print("HC-05 ready     ");
}

/**
  @brief  Konvertiert Dezimal oder Hex nach Int
  @param  Anzahl Nutzdaten (1 oder 2)
  @return Dezimalwert
*/
uint8_t ctoi(uint8_t value)
{
  if (value >= 'A')
    return value - 'A' + 10;
  else
    return value - '0';
}

/**
  @brief  Extrahiert die Nutzdaten 
          Antwort-String enthaelt Blank: "41 05 14" oder "41 0C D4 40"
  @param  Anzahl Nutzdaten (1 oder 2)
  @return Dezimalwert
*/
uint16_t extractOBDvalueDEC (uint8_t cnt)
{
  uint16_t value;

  value = (ctoi (str[6]) << 4) + ctoi (str[7]);       // Data: 1 byte
  if (cnt == 2)
    value = (value << 8) + (ctoi (str[9]) << 4) + ctoi (str[10]);     // 2 bytes  
  
  return value;
}

/**
  @brief  Berechnet den Datenwert
  @param  Bytewert, anzBytes (1 oder 2), Minimalwert, Maximalwert
  @return Dezimalwert
*/
double calcOBDvalue (uint16_t bytewert, uint8_t anzBytes, double minV, double maxV)
{
  double bits;

  if (anzBytes == 1)
    bits = 255.0;
  else
    bits = 65535.0;

  return ( ((double)bytewert/bits) * (maxV - minV) + minV );
}

void setup(){}      // useless

void loop()
{
  double value;
  
  pinMode(Pin_KEY, OUTPUT);       // this pin will pull the HC-05 pin 34 (key pin) HIGH to switch module to AT mode
  Serial.begin(115200);
  BTSerial.begin(38400);      // HC-05 default speed in AT command more

  lcd.init();                      // initialize the lcd 
  lcd.backlight();
  lcd.setCursor(0, 0);
  lcd.print("Elektor ELM327");

  initHC05();                 // initialize the HC-05

  lcd.setCursor(0, 1);
  lcd.print("Connect ELM ... ");

  sendOBDcmd ("");          // only \n => flush ELM
  sendOBDcmd ("ATZ");       // reset ELM
  sendOBDcmd ("ATE0");       // Echo Off
  sendOBDcmd ("ATSP0");       // Auto All protocols
  sendOBDcmd ("0100");      // Read available PIDs (connect)

  lcd.setCursor(0, 1);
  lcd.print("ELM OK          ");
  
  while(1)
  {
    sendOBDcmd ("0105");   
    Serial.print("Temp: ");
    value = calcOBDvalue (extractOBDvalueDEC(1), 1, -40, 215);
    Serial.print(value);   
    Serial.println("°C");
    lcd.home();
    lcd.setCursor(0, 0);
    lcd.print("Temp : ");
    lcd.print((int16_t)value);     // cast nach integer, da keine Nachkommastellen 
    lcd.print(" \337");       // octal 337 = ° (degree sign) on LCD - display maybe wrong char => delete this command
    lcd.print("C");
    lcd.print("    ");   // add some blanks to clear last output
   
    
    delay (200);
    sendOBDcmd ("010C");   
    Serial.print("Drehzahl: ");
    value = calcOBDvalue (extractOBDvalueDEC(2), 2, 0, 16383.75);
    Serial.print(value);   
    Serial.println(" upm");
    lcd.setCursor(0, 1);
    lcd.print("Motor: ");
    lcd.print((uint16_t)value);     // cast nach integer, da Nachkommastellen unsinnig sind
    lcd.print(" upm");
    lcd.print("   ");   // add some blanks to clear last output
    
    delay (500);    // Pause vor naechster Datenabfrage 
  }
}
